
__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import sys
import copy
import stat
import time
import subprocess

import radical.utils as ru

from ..  import utils     as rpu
from ..  import states    as rps
from ..  import constants as rpc


# ==============================================================================
#
class Agent(rpu.Worker):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        self.agent_name = cfg['agent_name']
        rpu.Worker.__init__(self, 'AgentWorker', cfg)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._log.debug('starting AgentWorker for %s' % self.agent_name)

        # everything which comes after the worker init is limited in scope to
        # the current process, and will not be available in the worker process.
        self._pilot_id    = self._cfg['pilot_id']
        self._session_id  = self._cfg['session_id']
        self.final_cause  = None

        # all components use the command channel for control messages
        self.declare_subscriber('command', rpc.AGENT_COMMAND_PUBSUB, self.command_cb)


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        # This callback is invoked as a thread in the process context of the
        # main agent (parent process) class.
        #
        # NOTE: That means it is *not* joined in the finalization of the run
        # loop (child), and the subscriber thread needs to be joined specifically in the
        # current process context.  At the moment that requires a call to
        # self._finalize() in the main process.

        cmd = msg['cmd']
        arg = msg['arg']

        self._log.info('agent command: %s %s' % (cmd, arg))

        if cmd == 'shutdown':

            # let agent know what caused the termination (first cause)
            if not self.final_cause:
                self.final_cause = arg

                self._log.info("shutdown command (%s)" % arg)
                self.stop()

            else:
                self._log.info("shutdown command (%s) - ignore" % arg)


    # --------------------------------------------------------------------------
    #
    def barrier_cb(self, topic, msg):

        # This callback is invoked in the process context of the run loop, and
        # will be cleaned up automatically.

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'alive':

            name = arg
            self._log.debug('waiting alive: \n%s\n%s\n%s'
                    % (self._components.keys(), self._workers.keys(),
                        self._sub_agents.keys()))

            # we only look at ALIVE messages which come from *this* agent, and
            # simply ignore all others (this is a shared medium after all)
            if name.startswith (self.agent_name):

                if name in self._components:
                    self._log.debug("component ALIVE (%s)" % name)
                    self._components[name]['alive'] = True

                elif name in self._workers:
                    self._log.debug("worker    ALIVE (%s)" % name)
                    self._workers[name]['alive'] = True

                else:
                    self._log.error("unknown   ALIVE (%s)" % name)

            elif name in self._sub_agents:
                self._log.debug("sub-agent ALIVE (%s)" % name)
                self._sub_agents[name]['alive'] = True


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):
        """
        Read the configuration file, setup logging and mongodb connection.
        This prepares the stage for the component setup (self._setup()).
        """

        # keep track of objects we need to stop in the finally clause
        self._sub_agents = dict()
        self._components = dict()
        self._workers    = dict()

        # sanity check on config settings
        if not 'cores'               in self._cfg: raise ValueError("Missing number of cores")
        if not 'debug'               in self._cfg: raise ValueError("Missing DEBUG level")
        if not 'lrms'                in self._cfg: raise ValueError("Missing LRMS")
        if not 'mongodb_url'         in self._cfg: raise ValueError("Missing MongoDB URL")
        if not 'pilot_id'            in self._cfg: raise ValueError("Missing pilot id")
        if not 'runtime'             in self._cfg: raise ValueError("Missing or zero agent runtime")
        if not 'scheduler'           in self._cfg: raise ValueError("Missing agent scheduler")
        if not 'session_id'          in self._cfg: raise ValueError("Missing session id")
        if not 'spawner'             in self._cfg: raise ValueError("Missing agent spawner")
        if not 'task_launch_method'  in self._cfg: raise ValueError("Missing unit launch method")
        if not 'agent_layout'        in self._cfg: raise ValueError("Missing agent layout")

        self._pilot_id   = self._cfg['pilot_id']
        self._session_id = self._cfg['session_id']
        self._runtime    = self._cfg['runtime']
        self._sub_cfg    = self._cfg['agent_layout'][self.agent_name]
        self._pull_units = self._sub_cfg.get('pull_units', False)

        # this better be on a shared FS!
        self._cfg['workdir'] = os.getcwd()

        # another sanity check
        if self.agent_name == 'agent_0':
            if self._sub_cfg.get('target', 'local') != 'local':
                raise ValueError("agent_0 must run on target 'local'")

        # configure the agent logger
        self._log.setLevel(self._cfg['debug'])

        # set up db connection -- only for the master agent and for the agent
        # which pulls units (which might be the same)
        if self.agent_name == 'agent_0' or self._pull_units:
            self._log.debug('connecting to mongodb at %s for unit pull')
            _, mongo_db, _, _, _  = ru.mongodb_connect(self._cfg['mongodb_url'])

            self._p  = mongo_db["%s.p"  % self._session_id]
            self._cu = mongo_db["%s.cu" % self._session_id]
            self._log.debug('connected to mongodb')

        # first order of business: set the start time and state of the pilot
        # Only the master agent performs this action
        if self.agent_name == 'agent_0':
            now = rpu.timestamp()
            ret = self._p.update(
                {"_id": self._pilot_id},
                {"$set" : {"state"        : rps.ACTIVE,
                           "started"      : now},
                 "$push": {"statehistory" : {"state"    : rps.ACTIVE,
                                             "timestamp": now}}
                })
            # TODO: Check for return value, update should be true!
            self._log.info("Database updated: %s", ret)

        # make sure we collect commands, specifically to implement the startup
        # barrier on bootstrap_4
        self.declare_publisher ('command', rpc.AGENT_COMMAND_PUBSUB)
        self.declare_subscriber('command', rpc.AGENT_COMMAND_PUBSUB, self.barrier_cb)

        # Now instantiate all communication and notification channels, and all
        # components and workers.  It will then feed a set of units to the
        # lead-in queue (staging_input).  A state notification callback will
        # then register all units which reached a final state (DONE).  Once all
        # units are accounted for, it will tear down all created objects.

        # we pick the layout according to our role (name)
        # NOTE: we don't do sanity checks on the agent layout (too lazy) -- but
        #       we would hiccup badly over ill-formatted or incomplete layouts...
        if not self.agent_name in self._cfg['agent_layout']:
            raise RuntimeError("no agent layout section for %s" % self.agent_name)

        try:
            self.start_sub_agents()
            self.start_components()

            # before we declare bootstrapping-success, the we wait for all
            # components, workers and sub_agents to complete startup.  For that,
            # all sub-agents will wait ALIVE messages on the COMMAND pubsub for
            # all entities it spawned.  Only when all are alive, we will
            # continue here.
            self.alive_barrier()

        except Exception as e:
            self._log.exception("Agent setup error: %s" % e)
            raise

        self._prof.prof('Agent setup done', logger=self._log.debug, uid=self._pilot_id)

        # also watch all components (once per second)
        self.declare_idle_cb(self.watcher_cb, 10.0)

        # once bootstrap_4 is done, we signal success to the parent agent
        # -- if we have any parent...
        if self.agent_name != 'agent_0':
            self.publish('command', {'cmd' : 'alive',
                                     'arg' : self.agent_name})

        # the pulling agent registers the staging_input_queue as this is what we want to push to
        # FIXME: do a sanity check on the config that only one agent pulls, as
        #        this is a non-atomic operation at this point
        self._log.debug('agent will pull units: %s' % bool(self._pull_units))
        if self._pull_units:

            self.declare_output(rps.AGENT_STAGING_INPUT_PENDING, rpc.AGENT_STAGING_INPUT_QUEUE)
            self.declare_publisher('state', rpc.AGENT_STATE_PUBSUB)

            # register idle callback, to pull for units -- which is the only action
            # we have to perform, really
            self.declare_idle_cb(self.idle_cb, self._cfg['db_poll_sleeptime'])


    # --------------------------------------------------------------------------
    #
    def alive_barrier(self):

        # FIXME: wait for bridges, too?  But we need pubsub for counting... Duh!
        total = len(self._components) + \
                len(self._workers   ) + \
                len(self._sub_agents)
        start   = time.time()
        timeout = 300

        while True:
            # check the procs for all components which are not yet alive
            to_check  = self._components.items() \
                      + self._workers.items() \
                      + self._sub_agents.items()

            alive_cnt = 0
            total_cnt = len(to_check)
            for name,c in to_check:
                if c['alive']:
                    alive_cnt += 1
                else:
                    self._log.debug('checking %s: %s', name, c)
                    if None != c['handle'].poll():
                        # process is dead and has never been alive.  Oops
                        raise RuntimeError('component %s did not come up' % name)

            self._log.debug('found alive: %2d / %2d' % (alive_cnt, total_cnt))

            if alive_cnt == total_cnt:
                self._log.debug('bootstrap barrier success')
                break

            if time.time() - timeout > start:
                raise RuntimeError('component barrier failed (timeout)')

            time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def watcher_cb(self):
        """
        we do a poll() on all our bridges, components, workers and sub-agent,
        to check if they are still alive.  If any goes AWOL, we will begin to
        tear down this agent.
        """

        to_watch = list(self._components.iteritems()) \
                 + list(self._workers.iteritems())    \
                 + list(self._sub_agents.iteritems())

      # self._log.debug('watch: %s' % pprint.pformat(to_watch))

        self._log.debug('checking %s things' % len(to_watch))
        for name, thing in to_watch:
            state = thing['handle'].poll()
            if state == None:
                self._log.debug('%-40s: ok' % name)
            else:
                raise RuntimeError ('%s died - shutting down' % name)

        return True # always idle


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        self._log.info("Agent finalizes")
        self._prof.prof('stop', uid=self._pilot_id)

        # tell other sub-agents get lost
        self.publish('command', {'cmd' : 'shutdown',
                                 'arg' : '%s finalization' % self.agent_name})

        # burn the bridges, burn EVERYTHING
        for name,sa in self._sub_agents.items():
            try:
                self._log.info("closing sub-agent %s", sa)
                sa['handle'].stop()
            except Exception as e:
                self._log.exception('ignore failing sub-agent terminate')

        for name,c in self._components.items():
            try:
                self._log.info("closing component %s", c)
                c['handle'].stop()
            except Exception as e:
                self._log.exception('ignore failing component terminate')

        for name,w in self._workers.items():
            try:
                self._log.info("closing worker %s", w)
                w['handle'].stop()
            except Exception as e:
                self._log.exception('ignore failing worker terminate')

        # communicate finalization to parent agent
        # -- if we have any parent...
        if self.agent_name != 'agent_0':
            self.publish('command', {'cmd' : 'final',
                                     'arg' : self.agent_name})

        self._log.info("Agent finalized")


    # --------------------------------------------------------------------------
    #
    def start_sub_agents(self):
        """
        For the list of sub_agents, get a launch command and launch that
        agent instance on the respective node.  We pass it to the seconds
        bootstrap level, there is no need to pass the first one again.
        """

        from ... import pilot as rp

        self._log.debug('start_sub_agents')

        sa_list = self._sub_cfg.get('sub_agents', [])

        if not sa_list:
            self._log.debug('start_sub_agents noop')
            return

        # the configs are written, and the sub-agents can be started.  To know
        # how to do that we create the agent launch method, have it creating
        # the respective command lines per agent instance, and run via
        # popen.
        #
        # actually, we only create the agent_lm once we really need it for
        # non-local sub_agents.
        agent_lm = None
        for sa in sa_list:
            target = self._cfg['agent_layout'][sa]['target']

            if target == 'local':

                # start agent locally
                cmdline = "/bin/sh -l %s/bootstrap_2.sh %s" % (os.getcwd(), sa)

            elif target == 'node':

                if not agent_lm:
                    agent_lm = rp.agent.LM.create(
                        name   = self._cfg['agent_launch_method'],
                        cfg    = self._cfg,
                        logger = self._log)

                node = self._cfg['lrms_info']['agent_nodes'][sa]
                # start agent remotely, use launch method
                # NOTE:  there is some implicit assumption that we can use
                #        the 'agent_node' string as 'agent_string:0' and
                #        obtain a well format slot...
                # FIXME: it is actually tricky to translate the agent_node
                #        into a viable 'opaque_slots' structure, as that is
                #        usually done by the schedulers.  So we leave that
                #        out for the moment, which will make this unable to
                #        work with a number of launch methods.  Can the
                #        offset computation be moved to the LRMS?
                # FIXME: are we using the 'hop' correctly?
                ls_name = "%s/%s.sh" % (os.getcwd(), sa)
                opaque_slots = {
                        'task_slots'   : ['%s:0' % node],
                        'task_offsets' : [],
                        'lm_info'      : self._cfg['lrms_info']['lm_info']}
                agent_cmd = {
                        'opaque_slots' : opaque_slots,
                        'description'  : {
                            'cores'      : 1,
                            'executable' : "/bin/sh",
                            'arguments'  : ["%s/bootstrap_2.sh" % os.getcwd(), sa]
                            }
                        }
                cmd, hop = agent_lm.construct_command(agent_cmd,
                        launch_script_hop='/usr/bin/env RP_SPAWNER_HOP=TRUE "%s"' % ls_name)

                with open (ls_name, 'w') as ls:
                    # note that 'exec' only makes sense if we don't add any
                    # commands (such as post-processing) after it.
                    ls.write('#!/bin/sh\n\n')
                    ls.write("exec %s\n" % cmd)
                    st = os.stat(ls_name)
                    os.chmod(ls_name, st.st_mode | stat.S_IEXEC)

                if hop : cmdline = hop
                else   : cmdline = ls_name

            # spawn the sub-agent
            self._prof.prof("create", msg=sa, uid=self._pilot_id)
            self._log.info ("create sub-agent %s: %s" % (sa, cmdline))
            sa_out = open("%s.out" % sa, "w")
            sa_err = open("%s.err" % sa, "w")
            sa_proc = subprocess.Popen(args=cmdline.split(), stdout=sa_out, stderr=sa_err)

            # make sure we can stop the sa_proc
            sa_proc.stop = sa_proc.terminate

            self._sub_agents[sa] = {'handle': sa_proc,
                                    'out'   : sa_out,
                                    'err'   : sa_err,
                                    'pid'   : sa_proc.pid,
                                    'alive' : False}
            self._prof.prof("created", msg=sa, uid=self._pilot_id)

        self._log.debug('start_sub_agents done')

    # --------------------------------------------------------------------------
    #
    def start_components(self):
        """
        For all componants defined on this agent instance, create the required
        number of those.  Keep a handle around for shutting them down later.
        """

        from ... import pilot as rp

        self._log.debug("start_components")

        # We use a static map from component names to class types for now --
        # a factory might be more appropriate (FIXME)
        cmap = {
            "AgentStagingInputComponent"  : rp.agent.Input,
            "AgentSchedulingComponent"    : rp.agent.Scheduler,
            "AgentExecutingComponent"     : rp.agent.Executing,
            "AgentStagingOutputComponent" : rp.agent.Output,
            }
        for cname, cnum in self._sub_cfg.get('components',{}).iteritems():
            for i in range(cnum):
                # each component gets its own copy of the config
                ccfg = copy.deepcopy(self._cfg)
                ccfg['number'] = i
                comp = cmap[cname].create(ccfg)
                comp.start()
                self._components[comp.childname] = {'handle' : comp,
                                                    'alive'  : False}
                self._log.info('created component %s (%s): %s', cname, cnum, comp.cname)

        # we also create *one* instance of every 'worker' type -- which are the
        # heartbeat and update worker.  To ensure this, we only create workers
        # in agent_0.
        # FIXME: make this configurable, both number and placement
        if self.agent_name == 'agent_0':
            wmap = {
                rpc.AGENT_UPDATE_WORKER    : rp.worker.Update,
                rpc.AGENT_HEARTBEAT_WORKER : rp.worker.Heartbeat
                }
            for wname in wmap:
                self._log.info('create worker %s', wname)
                wcfg   = copy.deepcopy(self._cfg)
                worker = wmap[wname].create(wcfg)
                worker.start()
                self._workers[worker.childname] = {'handle' : worker,
                                                   'alive'  : False}

        self._log.debug("start_components done")


    # --------------------------------------------------------------------------
    #
    def idle_cb(self):
        """
        This method will be driving all other agent components, in the sense
        that it will manage the connection to MongoDB to retrieve units, and
        then feed them to the respective component queues.
        """

        # only do something if configured to do so
        if not self._pull_units:
            self._log.debug('not configured to pull for units')
            return True  # fake work to avoid busy noops

        try:
            # check for new units
            return self.check_units()

        except Exception as e:
            # exception in the main loop is fatal
            self._log.exception("ERROR in agent main loop: %s" % e)
            sys.exit(1)


    # --------------------------------------------------------------------------
    #
    def check_units(self):

        # Check if there are compute units waiting for input staging
        # and log that we pulled it.
        #
        # FIXME: Unfortunately, 'find_and_modify' is not bulkable, so we have
        # to use 'find'.  To avoid finding the same units over and over again,
        # we update the state *before* running the next find -- so we do it
        # right here...  No idea how to avoid that roundtrip...
        # This also blocks us from using multiple ingest threads, or from doing
        # late binding by unit pull :/
        cu_cursor = self._cu.find(spec  = {"pilot"   : self._pilot_id,
                                           'state'   : rps.AGENT_STAGING_INPUT_PENDING,
                                           'control' : 'umgr'})
        if not cu_cursor.count():
            # no units whatsoever...
            self._log.info("units pulled:    0")
            return False

        # update the unit states to avoid pulling them again next time.
        cu_list = list(cu_cursor)
        cu_uids = [cu['_id'] for cu in cu_list]

        self._cu.update(multi    = True,
                        spec     = {"_id"   : {"$in"     : cu_uids}},
                        document = {"$set"  : {"control" : 'agent'}})

        self._log.info("units pulled: %4d"   % len(cu_list))
        self._prof.prof('get', msg="bulk size: %d" % len(cu_list), uid=self._pilot_id)
        for cu in cu_list:
            self._prof.prof('get', msg="bulk size: %d" % len(cu_list), uid=cu['_id'])

        # now we really own the CUs, and can start working on them (ie. push
        # them into the pipeline).  We don't publish nor profile as advance,
        # since that happened already on the module side when the state was set.
        self.advance(cu_list, publish=False, push=True, prof=False)

        # indicate that we did some work (if we did...)
        return True


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

from base import LRMS


# ==============================================================================
#
class SGE(LRMS):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LRMS.__init__(self, cfg, logger)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        sge_hostfile = os.environ.get('PE_HOSTFILE')
        if sge_hostfile is None:
            msg = "$PE_HOSTFILE not set!"
            self._log.error(msg)
            raise RuntimeError(msg)

        # SGE core configuration might be different than what multiprocessing
        # announces
        # Alternative: "qconf -sq all.q|awk '/^slots *[0-9]+$/{print $2}'"

        # Parse SGE hostfile for nodes
        sge_node_list = [line.split()[0] for line in open(sge_hostfile)]
        # Keep only unique nodes
        sge_nodes = list(set(sge_node_list))
        self._log.info("Found PE_HOSTFILE %s. Expanded to: %s", sge_hostfile, sge_nodes)

        # Parse SGE hostfile for cores
        sge_cores_count_list = [int(line.split()[1]) for line in open(sge_hostfile)]
        sge_core_counts = list(set(sge_cores_count_list))
        sge_cores_per_node = min(sge_core_counts)
        self._log.info("Found unique core counts: %s Using: %d", sge_core_counts, sge_cores_per_node)

        self.node_list = sge_nodes
        self.cores_per_node = sge_cores_per_node







import os
import sys
import time
import signal

import threading       as mt
import multiprocessing as mp
import radical.utils   as ru

from ..states    import *

from .prof_utils import Profiler, clone_units, drop_units
from .prof_utils import timestamp as util_timestamp

from .queue      import Queue        as rpu_Queue
from .queue      import QUEUE_ZMQ    as rpu_QUEUE_ZMQ
from .queue      import QUEUE_OUTPUT as rpu_QUEUE_OUTPUT
from .queue      import QUEUE_INPUT  as rpu_QUEUE_INPUT

from .pubsub     import Pubsub       as rpu_Pubsub
from .pubsub     import PUBSUB_ZMQ   as rpu_PUBSUB_ZMQ
from .pubsub     import PUBSUB_PUB   as rpu_PUBSUB_PUB
from .pubsub     import PUBSUB_SUB   as rpu_PUBSUB_SUB

# TODO:
#   - add PENDING states
#   - for notifications, change msg from [topic, unit] to [topic, msg]
#   - components should not need to declare the state publisher?


# ==============================================================================
#
class Component(mp.Process):
    """
    This class provides the basic structure for any RP component which operates
    on stateful units.  It provides means to:

      - define input channels on which to receive new units in certain states
      - define work methods which operate on the units to advance their state
      - define output channels to which to send the units after working on them
      - define notification channels over which messages with other components
        can be exchanged (publish/subscriber channels)

    All low level communication is handled by the base class -- deriving classes
    need only to declare the respective channels, valid state transitions, and
    work methods.  When a unit is received, the component is assumed to have
    full ownership over it, and that no other unit will change the unit state
    during that time.

    The main event loop of the component -- run() -- is executed as a separate
    process.  Components inheriting this class should be fully self sufficient,
    and should specifically attempt not to use shared resources.  That will
    ensure that multiple instances of the component can coexist, for higher
    overall system throughput.  Should access to shared resources be necessary,
    it will require some locking mechanism across process boundaries.

    This approach should ensure that

      - units are always in a well defined state;
      - components are simple and focus on the semantics of unit state
        progression;
      - no state races can occur on unit state progression;
      - only valid state transitions can be enacted (given correct declaration
        of the component's semantics);
      - the overall system is performant and scalable.

    Inheriting classes may overload the methods:

        initialize
        initialize_child
        finalize
        finalize_child

    These method should be used to
      - set up the component state for operation
      - declare input/output/notification channels
      - declare work methods
      - declare callbacks to be invoked on state notification
      - tear down the same on closing

    Inheriting classes MUST call the constructor:

        class StagingComponent(ComponentBase):
            def __init__(self, args):
                ComponentBase.__init__(self)

    Further, the class must implement the declared work methods, with
    a signature of:

        work(self, unit)

    The method is expected to change the unit state.  Units will not be pushed
    to outgoing channels automatically -- to do so, the work method has to call

        self.advance(unit)

    Until that method is called, the component is considered the sole owner of
    the unit.  After that method is called, the unit is considered disowned by
    the component.  It is the component's responsibility to call that method
    exactly once per unit.

    Having said that, components can return from the work methods without
    calling advance, for two reasons.

      - the unit may be in a final state, and is dropping out of the system (it
        will never again advance in the state model)
      - the component keeps ownership of the unit to advance it asynchronously
        at a later point in time.

    That implies that a component can collect ownership over an arbitrary number
    of units over time.  Either way, at most one work method instance will ever
    be active at any point in time.
    """

    # --------------------------------------------------------------------------
    #
    # FIXME:
    #  - *_PENDING -> * ?
    #  - make state transitions more formal
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    #
    def __init__(self, ctype, cfg):
        """
        This constructor MUST be called by inheriting classes.

        Note that __init__ is not executed in the process scope of the main
        event loop -- initialization for the main event loop should be moved to
        the initialize_child() method.  Initialization for component input,
        output and callbacks should be done in a separate initialize() method,
        to avoid the situation where __init__ creates threads but later fails
        and main thus ends up without a handle to terminate the threads (__del__
        can deadlock).  initialize() is called during start() in the parent's
        process context.
        """

        self._ctype         = ctype
        self._cfg           = cfg
        self._debug         = cfg.get('debug', 'DEBUG') # FIXME
        self._agent_name    = cfg['agent_name']
        self._cname         = "%s.%s.%d" % (self._agent_name, self._ctype, cfg.get('number', 0))
        self._childname     = "%s.child" % self._cname
        self._addr_map      = cfg['bridge_addresses']
        self._parent        = os.getpid() # pid of spawning process
        self._inputs        = list()      # queues to get units from
        self._outputs       = dict()      # queues to send units to
        self._publishers    = dict()      # channels to send notifications to
        self._subscribers   = list()      # callbacks for received notifications
        self._workers       = dict()      # where units get worked upon
        self._idlers        = list()      # idle_callback registry
        self._terminate     = mt.Event()  # signal for thread termination
        self._finalized     = False       # finalization guard
        self._is_parent     = None        # guard initialize/initialize_child
        self._exit_on_error = True        # FIXME: make configurable
        self._cb_lock       = mt.Lock()   # guard threaded callback invokations
        self._clone_cb      = None        # allocate resources on cloning units
        self._drop_cb       = None        # free resources on dropping clones

        # use agent_name for one log per agent, cname for one log per agent and component
        log_name = self._cname
        log_tgt  = self._cname + ".log"
        self._log = ru.get_logger(log_name, log_tgt, self._debug)
        self._log.info('creating %s' % self._cname)

        self._prof = Profiler(self._cname)

        # start the main event loop in a separate process.  At that point, the
        # component will basically detach itself from the parent process, and
        # will only maintain a handle to be used for shutdown
        mp.Process.__init__(self, name=self._cname)


    # --------------------------------------------------------------------------
    #
    @property
    def ctype(self):
        return self._ctype


    # --------------------------------------------------------------------------
    #
    @property
    def cname(self):
        return self._cname


    # --------------------------------------------------------------------------
    #
    @property
    def childname(self):
        return self._childname


    # --------------------------------------------------------------------------
    #
    def initialize(self):
        """
        This method may be overloaded by the components.  It is called *once* in
        the context of the parent process, upon start(), and should be used to
        set up component state before units arrive.
        """
        self._log.debug('base initialize (NOOP)')


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):
        """
        This method may be overloaded by the components.  It is called *once* in
        the context of the child process, upon start(), and should be used to
        set up component state before units arrive.
        """
        self._log.debug('base initialize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def finalize(self):
        """
        This method may be overloaded by the components.  It is called *once* in
        the context of the parent process, upon stop(), and should be used to
        tear down component state after units have been processed.
        """
        self._log.debug('base finalize (NOOP)')


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):
        """
        This method may be overloaded by the components.  It is called *once* in
        the context of the child process, upon stop(), and should be used to
        tear down component state after units have been processed.
        """
        self._log.debug('base finalize_child (NOOP)')


    # --------------------------------------------------------------------------
    #
    def start(self):
        """
        This method will start the child process.  *After* doing so, it will
        call the parent's initialize, so that this is only executed in the
        parent's process context (before fork).  Start will execute the run loop
        in the child process context, and in that context call
        initialize_child() before entering the loop.

        start() essentially performs:

            if fork():
                # parent
                initialize()
            else:
                # child
                initialize_child()
                run()

        """

        if self._is_parent != None:
            raise RuntimeError('start() can be called only once')

        self._is_parent = True # run will reset this for the child

        # make sure we don't keep any profile entries buffered across fork
        self._prof.flush()

        # fork child process
        mp.Process.start(self)

        try:
            # this is now the parent process context
            self.initialize()
        except Exception as e:
            self._log.exception ('initialize failed')
            self.stop()
            raise




    # --------------------------------------------------------------------------
    #
    def stop(self):
        """
        Shut down the process hosting the event loop.  If the parent calls
        stop(), the child is simply terminated (no child finalizers are
        called).  If the child calls stop() itself, child finalizers are called
        before calling exit.

        stop() can be called multiple times, and can be called from the
        MainThread, or from sub thread (such as callback invocations) -- but it
        should notes that, if called from a callback, it may not always be able
        to tear down all threads, specifically not the callback thread itself
        and the MainThread.  Safest is calling it once from each the parent's
        and child's MainThread.  Since the finalizers are only called on the
        first invocation to stop(), finalizers can happen in callback threads!

        stop() basically performs:

            tear down all subscriber threads
            if parent:
                finalize()
                self.terminate()
            else:
                finalize_child()
                sys.exit()
        """

        if self._finalized:
            # only die once
            return

        self._prof.prof("closing")
        self._log.info("closing (%d subscriber threads)" % (len(self._subscribers)))

        # tear down all subscriber threads
        self._terminate.set()
        self_thread = mt.current_thread()
        for t in self._subscribers:
            if t != self_thread:
                self._log.debug('joining  subscriber thread %s' % t)
                t.join()
            else:
                self._log.debug('skipping subscriber thread %s' % t)

        self._log.debug('subscriber threads joined')

        if self._is_parent:
            self._log.info("terminating")
            self._prof.prof("finalize")
            if not self._finalized:
                self._finalized = True
                self.finalize()
                self._prof.prof("finalized")
            else:
                self._prof.prof("not_yet_finalized")

            # Signal the child
            self._log.debug('signalling child')
            self.terminate()

            # Wait for the child process
            self._log.debug('waiting for child')
            self.join()
            self._log.debug('child done')

            self._prof.prof("stopped")
            self._prof.close()

            # If we are called from within a callback, that (means?) we will
            # have skipped one thread for joining above.
            #
            # Note that the thread is *not* joined at this point -- but the
            # parent should not block on shutdown anymore, as the thread is
            # at least gone.
            if self_thread in self._subscribers and self._cb_lock.locked():
                self._log.debug('release subscriber thread %s' % self_thread)
                sys.exit()

        else:
            # we only finalize in the child's main thread.
            # NOTE: this relies on us not to change the name of MainThread
            if self_thread.name == 'MainThread':
                if not self._finalized:
                    self._prof.prof("not_yet_finalized")
                    self._finalized = True
                    self._prof.prof("finalize")
                    self.finalize_child()
                    self._prof.prof("finalized")
                    self._prof.prof("stopped")
                    self._prof.close()
                else:
                    self._prof.prof("already_finalized - ERROR")

            # The child exits here.  If this call happens in a subscriber
            # thread, then it will be caught in the run loop of the main thread,
            # leading to the main thread's demize, which ends up here again...
            sys.exit()


    # --------------------------------------------------------------------------
    #
    def poll(self):
        """
        This is a wrapper around is_alive() which mimics the behavior of the same
        call in the subprocess.Popen class with the same name.  It does not
        return an exitcode though, but 'None' if the process is still
        alive, and always '0' otherwise
        """
        if self.is_alive():
            return None
        else:
            return 0


    # --------------------------------------------------------------------------
    #
    def declare_input(self, states, input):
        """
        Using this method, the component can be connected to a queue on which
        units are received to be worked upon.  The given set of states (which
        can be a single state or a list of states) will trigger an assert check
        upon unit arrival.

        For each state specified by an declare_input() call, the component MUST
        also declare an respective work method.
        """

        if not isinstance(states, list):
            states = [states]

        # get address for the queue
        addr = self._addr_map[input]['source']
        self._log.debug("using addr %s for input %s" % (addr, input))

        q = rpu_Queue.create(rpu_QUEUE_ZMQ, input, rpu_QUEUE_OUTPUT, addr)
        self._inputs.append([q, states])

        for state in states:
            self._log.debug('declared input     : %s : %s : %s' \
                    % (state, input, q.name))


    # --------------------------------------------------------------------------
    #
    def declare_output(self, states, output=None):
        """
        Using this method, the component can be connected to a queue to which
        units are sent after being worked upon.  The given set of states (which
        can be a single state or a list of states) will trigger an assert check
        upon unit departure.

        If a state but no output is specified, we assume that the state is
        final, and the unit is then considered 'dropped' on calling advance() on
        it.  The advance() will trigger a state notification though, and then
        mark the drop in the log.  No other component should ever again work on
        such a final unit.  It is the responsibility of the component to make
        sure that the unit is in fact in a final state.
        """

        if not isinstance(states, list):
            states = [states]

        for state in states:

            # we want a *unique* output queue for each state.
            if state in self._outputs:
                self._log.warn("%s replaces output for %s : %s -> %s" \
                        % (self._cname, state, self._outputs[state], output))

            if not output:
                # this indicates a final state
                self._outputs[state] = None
            else:
                # get address for the queue
                addr = self._addr_map[output]['sink']
                self._log.debug("using addr %s for output %s" % (addr, output))

                # non-final state, ie. we want a queue to push to
                q = rpu_Queue.create(rpu_QUEUE_ZMQ, output, rpu_QUEUE_INPUT, addr)
                self._outputs[state] = q

                self._log.debug('declared output    : %s : %s : %s' \
                     % (state, output, q.name))


    # --------------------------------------------------------------------------
    #
    def declare_worker(self, states, worker):
        """
        This method will associate a unit state with a specific worker.  Upon
        unit arrival, the unit state will be used to lookup the respective
        worker, and the unit will be handed over.  Workers should call
        self.advance(unit), in order to push the unit toward the next component.
        If, for some reason, that is not possible before the worker returns, the
        component will retain ownership of the unit, and should call advance()
        asynchronously at a later point in time.

        Worker invocation is synchronous, ie. the main event loop will only
        check for the next unit once the worker method returns.
        """

        if not isinstance(states, list):
            states = [states]

        # we want exactly one worker associated with a state -- but a worker can
        # be responsible for multiple states
        for state in states:
            if state in self._workers:
                self._log.warn("%s replaces worker for %s (%s)" \
                        % (self._cname, state, self._workers[state]))
            self._workers[state] = worker

            self._log.debug('declared worker    : %s : %s' \
                    % (state, worker.__name__))



    # --------------------------------------------------------------------------
    #
    def declare_idle_cb(self, cb, timeout=0.0):

        if None == timeout:
            timeout = 0.0
        timeout = float(timeout)

        self._idlers.append({
            'cb'      : cb,       # call this whenever we are idle
            'last'    : 0.0,      # was never called before
            'timeout' : timeout}) # call no more often than this many seconds

        self._log.debug('declared idler     : %s : %s' % (cb.__name__, timeout))


    # --------------------------------------------------------------------------
    #
    def declare_publisher(self, topic, pubsub):
        """
        Using this method, the compinent can declare certain notification topics
        (where topic is a string).  For each topic, a pub/sub network will be
        used to distribute the notifications to subscribers of that topic.

        The same topic can be sent to multiple channels -- but that is
        considered bad practice, and may trigger an error in later versions.
        """

        if topic not in self._publishers:
            self._publishers[topic] = list()

        # get address for pubsub
        addr = self._addr_map[pubsub]['sink']
        self._log.debug("using addr %s for pubsub %s" % (addr, pubsub))

        q = rpu_Pubsub.create(rpu_PUBSUB_ZMQ, pubsub, rpu_PUBSUB_PUB, addr)
        self._publishers[topic].append(q)

        self._log.debug('declared publisher : %s : %s : %s' \
                % (topic, pubsub, q.name))

        # FIXME: I am not exactly sure why this is 'needed', but declaring
        #        a published as above and then immediately publishing on that
        #        channel seems to sometimes lead to a loss of messages.
        time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def declare_subscriber(self, topic, pubsub, cb):
        """
        This method is complementary to the declare_publisher() above: it
        declares a subscription to a pubsub channel.  If a notification
        with matching topic is received, the registered callback will be
        invoked.  The callback MUST have the signature:

          callback(topic, msg)

        The subscription will be handled in a separate thread, which implies
        that the callback invocation will also happen in that thread.  It is the
        caller's responsibility to ensure thread safety during callback
        invocation.
        """

        # ----------------------------------------------------------------------
        def _subscriber(q, callback):
            try:
                while not self._terminate.is_set():
                    topic, msg = q.get_nowait(1000) # timout in ms
                    if topic and msg:
                        with self._cb_lock:
                            callback (topic=topic, msg=msg)
            except Exception as e:
                self._log.exception("subscriber failed")
                if self._exit_on_error:
                    raise
        # ----------------------------------------------------------------------

        # get address for pubsub
        addr = self._addr_map[pubsub]['source']
        self._log.debug("using addr %s for pubsub %s" % (addr, pubsub))

        # create a pubsub subscriber, and subscribe to the given topic
        q = rpu_Pubsub.create(rpu_PUBSUB_ZMQ, pubsub, rpu_PUBSUB_SUB, addr)
        q.subscribe(topic)

        t = mt.Thread(target=_subscriber, args=[q,cb],
                      name="%s.subscriber" % self.cname)
        t.start()
        self._subscribers.append(t)

        self._log.debug('%s declared subscriber: %s : %s : %s : %s' \
                % (self._cname, topic, pubsub, cb, t.name))

    # --------------------------------------------------------------------------
    #
    def declare_clone_cb(self, clone_cb):
        """
        The clone callback will be invoked whenever a unit is cloned after
        passing this component.  So, whenever the component allocates some
        resources for a unit which would conflict when being cloned, this
        callback allows to perform the required correction (hi scheduler!).
        """
        self._log.debug('declare clone_cb %s (%s)', clone_cb, os.getpid())
        self._clone_cb = clone_cb


    # --------------------------------------------------------------------------
    #
    def declare_drop_cb(self, drop_cb):
        """
        The drop callback will be invoked whenever a unit is dropped after
        passing this component.  So, whenever the component allocates some
        resources for a cloned unit which could otherwise not be reaped anymore,
        because the unit would not pass some reaping state or something, this
        callback allows to perform the required action (hi scheduler!).
        """
        self._log.debug('declare drop_cb %s (%s)', drop_cb, os.getpid())
        self._drop_cb = drop_cb


    # --------------------------------------------------------------------------
    #
    def _subscriber_check_cb(self):

        for t in self._subscribers:
            if not t.isAlive():
                raise RuntimeError('subscriber %s died' % t.name)


    # --------------------------------------------------------------------------
    #
    def run(self):
        """
        This is the main routine of the component, as it runs in the component
        process.  It will first initialize the component in the process context.
        Then it will attempt to get new units from all input queues
        (round-robin).  For each unit received, it will route that unit to the
        respective worker method.  Once the unit is worked upon, the next
        attempt ar getting a unit is up.
        """

        self._is_parent = False
        self._cname     = self.childname
        self._log       = ru.get_logger(self._cname, "%s.log" % self._cname, self._debug)
        self._prof      = Profiler(self._cname)

        # parent can call terminate, which we translate here into sys.exit(),
        # which is then excepted in the run loop below for an orderly shutdown.
        def sigterm_handler(signum, frame):
            sys.exit()
        signal.signal(signal.SIGTERM, sigterm_handler)

        # reset other signal handlers to their default
        signal.signal(signal.SIGINT,  signal.SIG_DFL)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

        # set process name
        try:
            import setproctitle as spt
            spt.setproctitle('radical.pilot %s' % self._cname)
        except Exception as e:
            pass


        try:
            # configure the component's logger
            log_name  = self._cname
            log_tgt   = self._cname + ".log"
            self._log = ru.get_logger(log_name, log_tgt, self._debug)
            self._log.info('running %s' % self._cname)

            # initialize_child() should declare all input and output channels, and all
            # workers and notification callbacks
            self._prof.prof('initialize')
            self.initialize_child()
            self._prof.prof('initialized')

            # register own idle callback to watch the subscriber threads (1/sec)
            self.declare_idle_cb(self._subscriber_check_cb, 1.0)

            # perform a sanity check: for each declared input state, we expect
            # a corresponding work method to be declared, too.
            for input, states in self._inputs:
                for state in states:
                    if not state in self._workers:
                        raise RuntimeError("%s: no worker declared for input state %s" \
                                        % self._cname, state)


            # The main event loop will repeatedly iterate over all input
            # channels, probing
            while not self._terminate.is_set():

                # if no action occurs in this iteration, invoke idle callbacks
                active = False

                # FIXME: for the default case where we have only one input
                #        channel, we can probably use a more efficient method to
                #        pull for units -- such as blocking recv() (with some
                #        timeout for eventual shutdown)
                # FIXME: a simple, 1-unit caching mechanism would likely remove
                #        the req/res overhead completely (for any non-trivial
                #        worker).
                for input, states in self._inputs:

                    # FIXME: the timeouts have a large effect on throughput, but
                    #        I am not yet sure how best to set them...
                    unit = input.get_nowait(1000) # timeout in microseconds
                    if not unit:
                        continue

                    state = unit['state']
                    uid   = unit['_id']

                    # assert that the unit is in an expected state
                    if state not in states:
                        self.advance(unit, FAILED, publish=True, push=False)
                        self._prof.prof(event='failed', msg="unexpected state %s" % state,
                                uid=unit['_id'], state=unit['state'], logger=self._log.error)
                        continue

                    # depending on the queue we got the unit from, we can either
                    # drop units or clone them to inject new ones
                    unit = drop_units(self._cfg, unit, self.ctype, 'input',
                                      drop_cb=self._drop_cb, logger=self._log)
                    if not unit:
                        self._prof.prof(event='drop', state=state,
                                uid=uid, msg=input.name)
                        continue

                    units = clone_units(self._cfg, unit, self.ctype, 'input',
                                        clone_cb=self._clone_cb, logger=self._log)

                    for _unit in units:

                        uid = _unit['_id']
                        active = True
                        self._prof.prof(event='get', state=state, uid=uid, msg=input.name)


                        # check if we have a suitable worker (this should always be
                        # the case, as per the assertion done before we started the
                        # main loop.  But, hey... :P
                        if not state in self._workers:
                            self._log.error("%s cannot handle state %s: %s" \
                                    % (self._cname, state, _unit))
                            continue

                        # we have an acceptable state and a matching worker -- hand
                        # it over, wait for completion, and then pull for the next
                        # unit
                        try:
                            self._prof.prof(event='work start', state=state, uid=uid)
                            with self._cb_lock:
                                self._workers[state](_unit)
                            self._prof.prof(event='work done ', state=state, uid=uid)

                        except Exception as e:
                            self.advance(_unit, FAILED, publish=True, push=False)
                            self._prof.prof(event='failed', msg=str(e), uid=uid, state=state)
                            self._log.exception("unit %s failed" % uid)

                            if self._exit_on_error:
                                raise


                # if nothing happened, we can call the idle callbacks.  Don't
                # call them more frequently than what they specified as sleep
                # time though!
                if not active:

                    now = time.time()

                    for idler in self._idlers:

                        if (now - idler['last']) > idler['timeout']:

                            try:
                                with self._cb_lock:
                                    if idler['cb']():
                                        # something happened!
                                        idler['last'] = now
                                        active = True
                            except Exception as e:
                                self._log.exception('idle cb failed')
                                if self._exit_on_error:
                                    raise

                if not active:
                    # FIXME: make configurable
                    time.sleep(0.1)


        except Exception as e:
            # We should see that exception only on process termination -- and of
            # course on incorrect implementations of component workers.  We
            # could in principle detect the latter within the loop -- - but
            # since we don't know what to do with the units it operated on, we
            # don't bother...
            self._log.exception('loop exception')

        except SystemExit:
            self._log.debug("Caught exit")

        except:
            # Can be any other signal or interrupt.
            self._log.exception('loop interruption')

        finally:
            # call stop (which calls the finalizers)
            self.stop()


    # --------------------------------------------------------------------------
    #
    def advance(self, units, state=None, publish=True, push=False, prof=True,
                timestamp=None):
        """
        Units which have been operated upon are pushed down into the queues
        again, only to be picked up by the next component, according to their
        state.  This method will update the unit state, and push it into the
        output queue declared as target for that state.

        units:   list of units to advance
        state:   new state to set for the units
        publish: determine if state update notifications should be issued
        push:    determine if units should be pushed to outputs
        prof:    determine if state advance creates a profile event
                 (publish, push, and drop are always profiled)
        """

        if not timestamp:
            timestamp = util_timestamp()

        if not isinstance(units, list):
            units = [units]

        for unit in units:

            uid = unit['_id']

            if state:
                unit['state']          = state
                unit['state_timstamp'] = timestamp
                if prof:
                    self._prof.prof('advance', uid=unit['_id'], state=state,
                            timestamp=timestamp)
            else:
                state = unit['state']

            if publish:
                # send state notifications
                self.publish('state', unit)
                self._prof.prof('publish', uid=unit['_id'], state=unit['state'])

            if push:
                if state not in self._outputs:
                    # unknown target state -- error
                    self._log.error("%s can't route state %s (%s)" \
                            % (self._cname, state, self._outputs.keys()))
                    continue

                if not self._outputs[state]:
                    # empty output -- drop unit
                    self._log.debug('%s %s ===| %s' % ('state', unit['id'], unit['state']))
                    continue


                output = self._outputs[state]

                # depending on the queue we got the unit from, we can now either
                # drop units or clone them to inject new ones
                unit = drop_units(self._cfg, unit, self.ctype, 'output',
                                  drop_cb=self._drop_cb, logger=self._log)
                if not unit:
                    self._prof.prof(event='drop', state=state, uid=uid, msg=output.name)
                    continue

                units = clone_units(self._cfg, unit, self.ctype, 'output',
                                    clone_cb=self._clone_cb, logger=self._log)

                for _unit in units:
                    # FIXME: we should assert that the unit is in a PENDING state.
                    #        Better yet, enact the *_PENDING transition right here...
                    #
                    # push the unit down the drain
                    output.put(_unit)
                    self._prof.prof('put', uid=_unit['_id'], state=state, msg=output.name)


    # --------------------------------------------------------------------------
    #
    def publish(self, topic, msg):
        """
        push information into a publication channel
        """

        if not isinstance(msg, list):
            msg = [msg]

        if topic not in self._publishers:
            self._log.error("%s can't route notification '%s:%s' (%s)" \
                    % (self._cname, topic, msg, self._publishers.keys()))
            return

        if not self._publishers[topic]:
            self._log.error("%s no route for notification '%s:%s' (%s)" \
                    % (self._cname, topic, msg, self._publishers.keys()))
            return

        for m in msg:
            for p in self._publishers[topic]:
                p.put (topic, m)



# ==============================================================================
#
class Worker(Component):
    """
    A Worker is a Component which cannot change the state of the unit it
    handles.  Workers are emplyed as helper classes to mediate between
    components, between components and database, and between components and
    notification channels.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, ctype, cfg):

        Component.__init__(self, ctype, cfg)


    # --------------------------------------------------------------------------
    #
    # we overload state changing methods from component and assert neutrality
    # FIXME: we should insert hooks around callback invocations, too
    #
    def advance(self, units, state=None, publish=True, push=False, prof=True):

        if state:
            raise RuntimeError("worker %s cannot advance state (%s)"
                    % (self.cname, state))

        Component.advance(self, units, state, publish, push, prof)



# ------------------------------------------------------------------------------




__author__    = "George Chantzialexiou"
__copyright__ = "Copyright  2014, The RADICAL Group"
__license__   = "MIT"

import os
import sys
import radical.pilot as rp
import math
import time


"""
    This is a simple implementation of k-means algorithm using the RADICAl-Pilot API.
"""
#------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):

    if not pilot:
        return

    print "[Callback]: ComputePilot '%s' state: %s." % (pilot.uid, state)

    if state == rp.FAILED:
        sys.exit (1)


#------------------------------------------------------------------------------
#
def unit_state_cb (unit, state):

    if not unit:
        return

    print "[Callback]: unit %s on %s: %s." % (unit.uid, unit.pilot_id, state)

    if state == rp.FAILED:
        print "stderr: %s" % unit.stderr
        # do not exit


#------------------------------------------------------------------------------
#
def wait_queue_size_cb(umgr, wait_queue_size):

    print "[Callback]: wait_queue_size: %s." % wait_queue_size


#------------------------------------------------------------------------------
#
def quickselect(x, k):    
    # k should start from 0
    # returns the element itself
     
    pivot = x[len(x) // 2]
    left = [e for e in x if e < pivot]
    right = [e for e in x if e > pivot]
    
    delta = len(x) - len(right)
    if k < len(left):
        return quickselect(left, k)
    elif k >= delta:
        return quickselect(right, k - delta)
    else:
        return pivot


#-------------------------------------------------------------------------
#
def get_distance(dataPointX, dataPointY, centroidX, centroidY):
    
    # Calculate Euclidean distance.
    return math.sqrt(math.pow((centroidY - dataPointY), 2) + math.pow((centroidX - dataPointX), 2))


#------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()
    print "session id: %s" % session.uid

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        start_time = time.time()

        # FUNCTIONS OF RADICAL PILOT
        # don't forget to change the localhost label
        c = rp.Context('ssh')
        #c.user_id = 'userid'
        session.add_context(c)
        
        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        print "Initiliazing Pilot Manager..."
        pmgr = rp.PilotManager(session=session)

        # Register our callback with our Pilot Manager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state
        pmgr.register_callback(pilot_state_cb)

        # This describes the requirements and the parameters
        pdesc = rp.ComputePilotDescription()
        pdesc.resource = "localhost"
        #pdesc.resource = "xsede.stampede"
        #pdesc.project = "TG-MCB090174"
        pdesc.runtime = 10 # minutes
        pdesc.cores = 4

        print "Submitting Compute Pilot to PilotManager"
        pilot = pmgr.submit_pilots(pdesc)

        # Combine all the units
        print "Initiliazing Unit Manager"

        # Combine the ComputePilot, the ComputeUnits and a scheduler via
        # a UnitManager object.
        umgr = rp.UnitManager(
            session=session,
            scheduler=rp.SCHED_DIRECT_SUBMISSION)

        # Register our callback with the UnitManager. This callback will get
        # called every time any of the units managed by the UnitManager
        # change their state.
        print 'Registering the callbacks so we can keep an eye on the CUs'
        umgr.register_callback(unit_state_cb)

        print "Registering Compute Pilot with Unit Manager"
        umgr.add_pilots(pilot) 
        # need to read the number of the k division the user want to do: that's also the only argument
        args = sys.argv[1:]
        if len(args) < 1:
            print "Usage: python %s needs the k-variable. Please run again" % __file__
            print "python k-means k"
            sys.exit(-1)
        k = int(sys.argv[1])  # k is the number of clusters i want to create
        # read the dataset from dataset.data file and pass the elements to x list
        try:
            data = open("dataset4.in",'r')
        except IOError:
            print "Missing dataset. file! Run:"
            sys.exit(-1)

        dataset_as_string_array = data.readline().split(',')
        x = map(float, dataset_as_string_array)
        data.close()
        #initialize the centroids   : choose the centroids and delete them from the element list (x)
        # FUNCTIONS OF RADICAL PILOT


        #  CHOOSING THE CENTROIDS
        centroid = []
        size = len(x)
        for i in range(0,k):
            a1 = quickselect(x,(size*(i+1))//((k+1)))
            centroid.append(a1)

        for i in range(0,k):
            x.remove(centroid[i])
        # END OF CHOOSING CENTROIDS
        
        # PUT THE CENTROIDS IN A FILE
        centroid_to_string = ','.join(map(str,centroid))
        centroid_file = open('centroids.txt', 'w')
        centroid_file.write(centroid_to_string)
        centroid_file.close()       
        # END OF PUTTING CENTROIDS IN A FILE CENTROIDS.DATA


        # VARIABLE DEFINITIONS
        p =   pdesc.cores  # NUMBER OF CORES OF THE SYSTEM I USE
        convergence = False   # We have no convergence yet
        m = 0 # number of iterations
        maxIt = 20 # the maximum number of iteration
        part_length = len(x)/p  # this is the length of the part that each unit is going to control
        # END OF VARIABLE DEFINITIONS


        while ((m<maxIt) and (convergence==False)):
            #  PUT THE CENTROIDS INTO DIFFERENT FILES
            for i in range(1,p+1):
                input_file = open("cu_%d.data" % i, "w")
                p1 = part_length*(i-1)
                p2 = part_length*(i) 
                if (i==p):
                    p2 = len(x)
                input_string = ','.join(map(str,x[p1:p2]))
                input_file.write(input_string)
                input_file.close()
            # END OF PUTTING CENTROIDS INTO DIFFERENT FILES
            
            
            # GIVE THE FILES INTO CUS TO START CALCULATING THE CLUSTERS - PHASE A
            mylist = []
            for i in range(1,p+1):
                cudesc = rp.ComputeUnitDescription()    
                cudesc.executable  = "python"
                cudesc.arguments = ['clustering_the_elements.py', i, k]
                cudesc.input_staging = ['clustering_the_elements.py', 'cu_%d.data' % i,'centroids.txt']
                cudesc.output_staging = 'centroid_cu_%d.data' % i
                mylist.append(cudesc)
                
            print 'Submitting the CU to the Unit Manager...'
            mylist_units = umgr.submit_units(mylist)
            # wait for all units to finish
            umgr.wait_units()
            print "All Compute Units completed PhaseA successfully! Now.." 
            #print " We compose the files into k files "

            # END OF PHASE A - NOW WE HAVE THE CLUSTERS CALCULATED IN THE CU FILES


            # FINDING THE AVG_ELEMENTS WHICH ARE CANDIDATE CENTROIDS 

            sum_of_all_centroids = []
            for i in range(0,2*k):
                sum_of_all_centroids.append(0)
            
            for i in range(1,p+1):
                read_file = open("centroid_cu_%d.data" % i, "r")
                read_as_string_array = read_file.readline().split(',')
                read_as_float_array = map(float, read_as_string_array)
                for j in range(0,k):
                    sum_of_all_centroids[2*j] += read_as_float_array[2*j]
                    sum_of_all_centroids[(2*j)+1] += read_as_float_array[(2*j)+1]
                read_file.close()

            for i in range(0,k):
                if (sum_of_all_centroids[(2*i)+1] != 0):
                    sum_of_all_centroids[i] = sum_of_all_centroids[(2*i)]/sum_of_all_centroids[(2*i)+1]
                else:
                    sum_of_all_centroids[i] = -1   #there are no centroids in this cluster

            # writing the elements to a file
            input_file = open('centroids.txt','w')
            input_string = ','.join(map(str,sum_of_all_centroids[0:p])) 
            input_file.write(input_string)
            input_file.close()


            # END THE AVG_ELEMENTS WHICH ARE CANDIATE CENTROIDS
            mylist = []

            for i in range(1,p+1):
                cudesc = rp.ComputeUnitDescription()    
                cudesc.executable  = "python"
                cudesc.arguments = ['finding_the_new_centroids.py', i, k]
                cudesc.input_staging = ['finding_the_new_centroids.py', 'cu_%d.data' % i,'centroids.txt']
                cudesc.output_staging = 'centroid_cu_%d.data' % i
                mylist.append(cudesc)
                
            print 'Submitting the CU to the Unit Manager...'
            mylist_units = umgr.submit_units(mylist)
            # wait for all units to finish
            umgr.wait_units()
            print "All Compute Units completed PhaseB successfully! Now.." 
            #print " We compose the files into k files "

            
            # FINDING THE NEW CENTROIDS- THESE ARE THE ELEMENTS WHO ARE CLOSER TO THE AVG_ELEMENTS
                # sum_of_all_centroids have the avg_elemnt
            new_centroids = []
            a = sys.maxint
            for i in range(0,k):
                new_centroids.append(a)
            
            for i in range(1,p+1):
                read_file = open("centroid_cu_%d.data" % i, "r")
                read_as_string_array = read_file.readline().split(',')
                read_as_float_array = map(float,read_as_string_array)
                for j in range(0,k):
                    if get_distance(new_centroids[j],0,sum_of_all_centroids[j],0) > get_distance(read_as_float_array[j],0,sum_of_all_centroids[j],0):
                        new_centroids[j] = read_as_float_array[j]
                read_file.close()


            # END OF FINDING THE NEW CENTROIDS- THESE ARE THE ELEMENTS WHO ARE CLOSER TO THE AVG_ELEMENT
         

            # CHECKING FOR CONVERGENCE
            
            # now we check for convergence - the prev centroids are in !centroid! and the new are in !new_centroids! 
            print "Now we check the converge"
            print 'new centroids:'
            print new_centroids
            print 'Old centroids:'
            print centroid
           # print 'element list'
           # print x
            convergence = True
            for i in range(0,len(new_centroids)):
                if (abs(new_centroids[i] - centroid[i]) > 1000): # i have elements from 1 to 10.000 so I give 1000 convergence
                    convergence = False

            if (convergence == False):
                m +=1
                for l in range(0,k):   # put the old centroids to the element file... remove the new cetroids from element list
                    if (centroid[l] != new_centroids[l]):
                        x.remove(new_centroids[l])
                        x.append(centroid[l])
                centroid =new_centroids
                input_string = ','.join(map(str,new_centroids))
                input_file = open('centroids.txt', 'w')
                input_file.write(input_string)
                input_file.close()
            
            # END OF CHECKING FOR CONVERGENCE

        print 'K-means algorithm ended successfully after %d iterations' % m
        # the only thing i have to do here is to check for convergence & add the file- arguments into the enviroments
        #print 'Thre centroids are in the cetroidss.txt file, and the elements of each centroid at centroid_x.data file'
        print 'The centroids of these elements are: \n'
        print new_centroids
        
        # END OF K-MEANS ALGORITHM
        finish_time = time.time()
        total_time = finish_time - start_time  # total execution time
        print 'The total execution time is: %f seconds' % total_time
        total_time /= 60

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print "caught Exception: %s" % e
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print "need to exit now: %s" % e

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print "closing session"
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


#-------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


from .base import LaunchMethod

# ==============================================================================
#
class CCMRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LaunchMethod.__init__(self, cfg, logger)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        # ccmrun: Cluster Compatibility Mode (CCM) job launcher for Cray systems
        self.launch_command= self._which('ccmrun')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        opaque_slots = cu['opaque_slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cores']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        ccmrun_command = "%s -n %d %s" % (self.launch_command, task_cores, task_command)

        return ccmrun_command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import subprocess
import multiprocessing

from base import LRMS


# ==============================================================================
#
class Yarn(LRMS):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LRMS.__init__(self, cfg, logger)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self._log.info("Using YARN on localhost.")

        selected_cpus = self.requested_cores

        # when we profile the agent, we fake any number of cores, so don't
        # perform any sanity checks.  Otherwise we use at most all available
        # cores (and informa about unused ones)
        if 'RADICAL_PILOT_PROFILE' not in os.environ:

            detected_cpus = multiprocessing.cpu_count()

            if detected_cpus < selected_cpus:
                self._log.warn("insufficient cores: using available %d instead of requested %d.",
                        detected_cpus, selected_cpus)
                selected_cpus = detected_cpus

            elif detected_cpus > selected_cpus:
                self._log.warn("more cores available: using requested %d instead of available %d.",
                        selected_cpus, detected_cpus)

        hdfs_conf_output =subprocess.check_output(['hdfs', 'getconf', 
                                                   '-nnRpcAddresses']).split('\n')[0]
        self._log.debug('Namenode URL = {0}'.format(hdfs_conf_output))
        self.namenode_url = hdfs_conf_output


        self._log.debug('Namenode URL = {0}'.format(self.namenode_url))

        # I will leave it for the moment because I have not found another way
        # to take the necessary value yet.
        yarn_conf_output = subprocess.check_output(['yarn', 'node', '-list'], 
                                                   stderr=subprocess.STDOUT).split('\n')
        for line in yarn_conf_output:
            if 'ResourceManager' in line:
                settings = line.split('at ')[1]
                if '/' in settings:
                    rm_url=settings.split('/')[1]
                    self.rm_ip=rm_url.split(':')[0]
                    self.rm_port=rm_url.split(':')[1]

                else:
                    self.rm_ip=settings.split(':')[0]
                    self.rm_port=settings.split(':')[1]

        hostname = os.environ.get('HOSTNAME')

        if hostname == None:
            self.node_list = ['localhost']
        else:
            self.node_list = [hostname]
        self.cores_per_node = selected_cpus







__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import threading

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
# 'enum' for RPs's pilot scheduler types
SCHEDULER_NAME_CONTINUOUS   = "CONTINUOUS"
SCHEDULER_NAME_SCATTERED    = "SCATTERED"
SCHEDULER_NAME_TORUS        = "TORUS"
SCHEDULER_NAME_YARN         = "YARN"


# ==============================================================================
#
class AgentSchedulingComponent(rpu.Component):

    # FIXME: clarify what can be overloaded by Scheduler classes

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        self._slots = None
        self._lrms  = None

        rpu.Component.__init__(self, rpc.AGENT_SCHEDULING_COMPONENT, cfg)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

      # self.declare_input (rps.AGENT_SCHEDULING_PENDING, rpc.AGENT_SCHEDULING_QUEUE)
      # self.declare_worker(rps.AGENT_SCHEDULING_PENDING, self.work)

        self.declare_input (rps.ALLOCATING_PENDING, rpc.AGENT_SCHEDULING_QUEUE)
        self.declare_worker(rps.ALLOCATING_PENDING, self.work)

        self.declare_output(rps.EXECUTING_PENDING,  rpc.AGENT_EXECUTING_QUEUE)

        # we need unschedule updates to learn about units which free their
        # allocated cores.  Those updates need to be issued after execution, ie.
        # by the AgentExecutionComponent.
        self.declare_publisher ('state',      rpc.AGENT_STATE_PUBSUB)
        self.declare_subscriber('unschedule', rpc.AGENT_UNSCHEDULE_PUBSUB, self.unschedule_cb)

        # we create a pubsub pair for reschedule trigger
        self.declare_publisher ('reschedule', rpc.AGENT_RESCHEDULE_PUBSUB)
        self.declare_subscriber('reschedule', rpc.AGENT_RESCHEDULE_PUBSUB, self.reschedule_cb)

        # all components use the command channel for control messages
        self.declare_publisher ('command',    rpc.AGENT_COMMAND_PUBSUB)

        # we declare a clone and a drop callback, so that cores can be assigned
        # to clones, and can also be freed again.
        self.declare_clone_cb(self.clone_cb)
        self.declare_drop_cb (self.drop_cb)

        # when cloning, we fake scheduling via round robin over all cores.
        # These indexes keeps track of the last used core.
        self._clone_slot_idx = 0
        self._clone_core_idx = 0

        # The scheduler needs the LRMS information which have been collected
        # during agent startup.  We dig them out of the config at this point.
        self._pilot_id = self._cfg['pilot_id']
        self._lrms_info           = self._cfg['lrms_info']
        self._lrms_lm_info        = self._cfg['lrms_info']['lm_info']
        self._lrms_node_list      = self._cfg['lrms_info']['node_list']
        self._lrms_cores_per_node = self._cfg['lrms_info']['cores_per_node']
        # FIXME: this information is insufficient for the torus scheduler!

        self._wait_pool = list()            # set of units which wait for the resource
        self._wait_lock = threading.RLock() # look on the above set
        self._slot_lock = threading.RLock() # look for slot allocation/deallocation

        # configure the scheduler instance
        self._configure()

        # communicate successful startup
        self.publish('command', {'cmd' : 'alive',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        # communicate finalization
        self.publish('command', {'cmd' : 'final',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Scheduler.
    #
    @classmethod
    def create(cls, cfg):

        # Make sure that we are the base-class!
        if cls != AgentSchedulingComponent:
            raise TypeError("Scheduler Factory only available to base class!")

        name = cfg['scheduler']

        from .continuous import Continuous
        from .scattered  import Scattered
        from .torus      import Torus
        from .yarn       import Yarn

        try:
            impl = {
                SCHEDULER_NAME_CONTINUOUS : Continuous,
                SCHEDULER_NAME_SCATTERED  : Scattered,
                SCHEDULER_NAME_TORUS      : Torus,
                SCHEDULER_NAME_YARN       : Yarn
            }[name]

            impl = impl(cfg)
            return impl

        except KeyError:
            raise ValueError("Scheduler '%s' unknown or defunct" % name)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        raise NotImplementedError("_configure() not implemented for Scheduler '%s'." % self._cname)


    # --------------------------------------------------------------------------
    #
    def slot_status(self):
        raise NotImplementedError("slot_status() not implemented for Scheduler '%s'." % self._cname)


    # --------------------------------------------------------------------------
    #
    def _allocate_slot(self, cores_requested):
        raise NotImplementedError("_allocate_slot() not implemented for Scheduler '%s'." % self._cname)


    # --------------------------------------------------------------------------
    #
    def _release_slot(self, opaque_slots):
        raise NotImplementedError("_release_slot() not implemented for Scheduler '%s'." % self._cname)


    # --------------------------------------------------------------------------
    #
    def _try_allocation(self, cu):
        """
        Attempt to allocate cores for a specific CU.  If it succeeds, send the
        CU off to the ExecutionWorker.
        """

        # needs to be locked as we try to acquire slots, but slots are freed
        # in a different thread.  But we keep the lock duration short...
        with self._slot_lock :

            # schedule this unit, and receive an opaque handle that has meaning to
            # the LRMS, Scheduler and LaunchMethod.
            cu['opaque_slots'] = self._allocate_slot(cu['description']['cores'])

        if not cu['opaque_slots']:
            # signal the CU remains unhandled
            return False

        # got an allocation, go off and launch the process
        self._prof.prof('schedule', msg="allocated", uid=cu['_id'])
        self._log.info("slot status after allocated  : %s" % self.slot_status ())

        return True


    # --------------------------------------------------------------------------
    #
    def reschedule_cb(self, topic, msg):
        # we ignore any passed CU.  In principle the cu info could be used to
        # determine which slots have been freed.  No need for that optimization
        # right now.  This will become interesting once reschedule becomes too
        # expensive.

        cu = msg

        self._prof.prof('reschedule', uid=self._pilot_id)
        self._log.info("slot status before reschedule: %s" % self.slot_status())

        # cycle through wait queue, and see if we get anything running now.  We
        # cycle over a copy of the list, so that we can modify the list on the
        # fly
        for cu in self._wait_pool[:]:

            if self._try_allocation(cu):

                # allocated cu -- advance it
                self.advance(cu, rps.EXECUTING_PENDING, publish=True, push=True)

                # remove it from the wait queue
                with self._wait_lock :
                    self._wait_pool.remove(cu)
                    self._prof.prof('unqueue', msg="re-allocation done", uid=cu['_id'])
            else:
                # Break out of this loop if we didn't manage to schedule a task
                break

        # Note: The extra space below is for visual alignment
        self._log.info("slot status after  reschedule: %s" % self.slot_status ())
        self._prof.prof('reschedule done')


    # --------------------------------------------------------------------------
    #
    def unschedule_cb(self, topic, msg):
        """
        release (for whatever reason) all slots allocated to this CU
        """

        cu = msg
        self._prof.prof('unschedule', uid=cu['_id'])

        if not cu['opaque_slots']:
            # Nothing to do -- how come?
            self._log.warn("cannot unschedule: %s (no slots)" % cu)
            return

        self._log.info("slot status before unschedule: %s" % self.slot_status ())

        # needs to be locked as we try to release slots, but slots are acquired
        # in a different thread....
        with self._slot_lock :
            self._release_slot(cu['opaque_slots'])
            self._prof.prof('unschedule', msg='released', uid=cu['_id'])

        # notify the scheduling thread, ie. trigger a reschedule to utilize
        # the freed slots
        self.publish('reschedule', cu)

        # Note: The extra space below is for visual alignment
        self._log.info("slot status after  unschedule: %s" % self.slot_status ())


    # --------------------------------------------------------------------------
    #
    def clone_cb(self, unit, name=None, mode=None, prof=None, logger=None):

        if mode == 'output':

            # so, this is tricky: we want to clone the unit after scheduling,
            # but at the same time don't want to have all clones end up on the
            # same core -- so the clones should be scheduled to a different (set
            # of) core(s).  But also, we don't really want to schedule, that is
            # why we blow up on output, right?
            #
            # So we fake scheduling.  This assumes the 'self._slots' structure as
            # used by the continuous scheduler, wo will likely only work for
            # this one (FIXME): we walk our own index into the slot structure,
            # and simply assign that core, be it busy or not.
            #
            # FIXME: This method makes no attempt to set 'task_slots', so will
            # not work properly for some launch methods.
            #
            # This is awful.  I mean, really awful.  Like, nothing good can come
            # out of this.  Ticket #902 should be implemented, it will solve
            # this problem much cleaner...

            if prof: prof.prof      ('clone_cb', uid=unit['_id'])
            else   : self._prof.prof('clone_cb', uid=unit['_id'])

            slot = self._slots[self._clone_slot_idx]

            unit['opaque_slots']['task_slots'][0] = '%s:%d' \
                    % (slot['node'], self._clone_core_idx)
          # self._log.debug(' === clone cb out : %s', unit['opaque_slots'])

            if (self._clone_core_idx +  1) < self._lrms_cores_per_node:
                self._clone_core_idx += 1
            else:
                self._clone_core_idx  = 0
                self._clone_slot_idx += 1

                if self._clone_slot_idx >= len(self._slots):
                    self._clone_slot_idx = 0


    # --------------------------------------------------------------------------
    #
    def drop_cb(self, unit, name=None, mode=None, prof=None, logger=None):

        if mode == 'output':
            # we only unscheduler *after* scheduling.  Duh!

            if prof:
                prof.prof('drop_cb', uid=unit['_id'])
            else:
                self._prof.prof('drop_cb', uid=unit['_id'])

            self.unschedule_cb(topic=None, msg=unit)



    # --------------------------------------------------------------------------
    #
    def work(self, cu):

      # self.advance(cu, rps.AGENT_SCHEDULING, publish=True, push=False)
        self.advance(cu, rps.ALLOCATING      , publish=True, push=False)

        # we got a new unit to schedule.  Either we can place it
        # straight away and move it to execution, or we have to
        # put it on the wait queue.
        if self._try_allocation(cu):
            self._prof.prof('schedule', msg="allocation succeeded", uid=cu['_id'])
            self.advance(cu, rps.EXECUTING_PENDING, publish=True, push=True)

        else:
            # No resources available, put in wait queue
            self._prof.prof('schedule', msg="allocation failed", uid=cu['_id'])
            with self._wait_lock :
                self._wait_pool.append(cu)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "George Chantzialexiou"
__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


""" A Simple Mandelbrot Fractal Generator.

    We use this example to explore the distributed capabilities of
    the Pilot Job and Filesystem APIs. The mandelbrot module
    calculates  partial mandelbrot set fractal and writes it to a 
    PNG image file. To determine which part of the fractal we create
    we use the index variable.

    It requires the Python Image Library (PIL) which can be easily
    installed with 'pip install Pillow'.

    This program is required by the mandelbrot_pilot.py program
    which creates the mandelbrot fractal using the capabilities of
    Pilot Job API. 

    On the command line:

        python mandelbrot.py imgX imgY xBeg xEnd yBeg yEnd cores index

    The parameters are as follows:

        imgX, imgY: the dimensions of the mandelbrot image, e.g. 1024, 1024
        xBeg, xEnd: the x-axis portion of the (sub-)image to calculate
        yBeg, yEnd: the y-axis portion of the (sub-)image to calculate
        cores : the number of the system's cores
        index: the part of the image we create. Can vary from 1 to cores.
"""

import sys
from PIL import Image

# ------------------------------------------------------------------------------
#
def makemandel(x_pixel_max, y_pixel_max, x_real_beg, x_real_end, y_real_beg, y_real_end, cores, index):

    maxIt  = 128 
    image  = Image.new("RGB", (x_pixel_max, y_pixel_max))

    y_real_range   = y_real_end - y_real_beg
    y_real_slice   = float(y_real_range) / float(cores)
    y_real_offset  = float(y_real_slice) * float(index)

    y_pixel_slice  = int(y_pixel_max / cores)
    y_pixel_offset = y_pixel_slice * index

    x_real_range   = x_real_end - x_real_beg
    x_real_slice   = float(x_real_range)
    x_pixel_slice  = int(x_pixel_max)
    x_pixel_offset = 0

    for y in range(y_pixel_slice):
        cy = y_real_beg + y_real_offset + (y_real_slice/y_pixel_slice*y)
        for x in range(x_pixel_slice):
            cx = x_real_beg + (x_real_slice/x_pixel_slice*x)

            c = complex(cx, cy)
            z = 0
            for i in range(maxIt):
                if abs(z) > 2.0: break 
                z = z * z + c 
            r = i % 4 * 16
            g = i % 6 * 16
            b = i % 16 * 16
            image.putpixel((x, y_pixel_slice*index+y), b * 65536 + g * 256 + r)
 
    image.save('mandel_%d.gif' % index , "GIF")
    
    return image

# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    args = sys.argv[1:]

    imgX  =   int(sys.argv[1])
    imgY  =   int(sys.argv[2])
    xBeg  = float(sys.argv[3])
    xEnd  = float(sys.argv[4])
    yBeg  = float(sys.argv[5])
    yEnd  = float(sys.argv[6])
    cores =   int(sys.argv[7])
    index =   int(sys.argv[8])

    makemandel(imgX, imgY, xBeg, xEnd, yBeg, yEnd, cores, index)
    sys.exit(0)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

from .base import LaunchMethod


# ==============================================================================
#
class SSH(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LaunchMethod.__init__(self, cfg, logger)

        # Instruct the ExecWorkers to unset this environment variable.
        # Otherwise this will break nested SSH with SHELL spawner, i.e. when
        # both the sub-agent and CUs are started using SSH.
        self.env_removables.extend(["RP_SPAWNER_HOP"])


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        # Find ssh command
        command = self._which('ssh')

        if command is not None:

            # Some MPI environments (e.g. SGE) put a link to rsh as "ssh" into
            # the path.  We try to detect that and then use different arguments.
            if os.path.islink(command):

                target = os.path.realpath(command)

                if os.path.basename(target) == 'rsh':
                    self._log.info('Detected that "ssh" is a link to "rsh".')
                    return target

            command = '%s -o StrictHostKeyChecking=no' % command

        self.launch_command = command


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        opaque_slots = cu['opaque_slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cores']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if not 'task_slots' in opaque_slots:
            raise RuntimeError('insufficient information to launch via %s: %s' \
                    % (self.name, opaque_slots))

        task_slots = opaque_slots['task_slots']

        if not launch_script_hop :
            raise ValueError ("LaunchMethodSSH.construct_command needs launch_script_hop!")

        # Get the host of the first entry in the acquired slot
        host = task_slots[0].split(':')[0]

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        # Pass configured and available environment variables to the remote shell
        export_vars = ' '.join(['%s=%s' % (var, os.environ[var]) for var in self.EXPORT_ENV_VARIABLES if var in os.environ])

        # Command line to execute launch script via ssh on host
        ssh_hop_cmd = "%s %s %s %s" % (self.launch_command, host, export_vars, launch_script_hop)

        # Special case, return a tuple that overrides the default command line.
        return task_command, ssh_hop_cmd


# ------------------------------------------------------------------------------




__author__ = 'vivek'
import os
import sys
from extasy import script

if __name__ == '__main__':
    print 'creating new crd files...'
    nreps = int(sys.argv[1])
    cycle = int(sys.argv[2])
    dict = {}
    dict['cycle'] = cycle
    for rep in range(nreps):
        tl = script.Script()
        tl.append('source leaprc.ff99SB')
        tl.append('x = loadpdb pdbs/%s.pdb'%(rep))
        tl.append('saveamberparm x delete.me min%s%s.crd'%(cycle,rep))
        tl.append('quit')
        tl.run('tleap -f {}')



#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

os.environ['RADICAL_PILOT_VERBOSE'] = 'REPORT'
os.environ['RADICAL_PILOT_PROFILE'] = 'TRUE'

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
# READ the RADICAL-Pilot documentation: http://radicalpilot.readthedocs.org/
#
# ------------------------------------------------------------------------------


#------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # we use a reporter class for nicer output
    report = ru.LogReporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv)  > 2: report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2: resource = sys.argv[1]
    else                   : resource = 'local.localhost'

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/../config.json' % os.path.dirname(os.path.abspath(__file__)))
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        pmgr = rp.PilotManager(session=session)

        # Define an [n]-core local pilot that runs for [x] minutes
        # Here we use a dict to initialize the description object
        report.info('create pilot description')
        pd_init = {
                'resource'      : resource,
                'cores'         : 64,  # pilot size
                'runtime'       : 15,  # pilot runtime (min)
                'exit_on_error' : True,
                'project'       : config[resource]['project'],
                'queue'         : config[resource]['queue'],
                'access_schema' : config[resource]['schema']
                }
        pdesc = rp.ComputePilotDescription(pd_init)
        report.ok('>>ok\n')

        # Launch the pilot.
        pilot = pmgr.submit_pilots(pdesc)


        report.header('submit units')

        # Register the ComputePilot in a UnitManager object.
        umgr = rp.UnitManager(session=session)
        umgr.add_pilots(pilot)

        # Create a workload of ComputeUnits.
        # Each compute unit runs '/bin/date'.

        n = 128   # number of units to run
        report.info('create %d unit description(s)\n\t' % n)

        cuds = list()
        for i in range(0, n):

            # create a new CU description, and fill it.
            # Here we don't use dict initialization.
            cud = rp.ComputeUnitDescription()
            # trigger an error now and then
            if not i % 10: cud.executable = '/bin/data' # does not exist
            else         : cud.executable = '/bin/date'

            cuds.append(cud)
            report.progress()
        report.ok('>>ok\n')

        # Submit the previously created ComputeUnit descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning ComputeUnits to the ComputePilots.
        units = umgr.submit_units(cuds)

        # Wait for all compute units to reach a final state (DONE, CANCELED or FAILED).
        report.header('gather results')
        umgr.wait_units()
    
        report.info('\n')
        for unit in units:
            if unit.state == rp.FAILED:
                report.plain('  * %s: %s, exit: %3s, err: %s' \
                        % (unit.uid, unit.state[:4], 
                           unit.exit_code, unit.stderr.strip()[-35:]))
                report.error('>>err\n')
            else:
                report.plain('  * %s: %s, exit: %3s, out: %s' \
                        % (unit.uid, unit.state[:4], 
                            unit.exit_code, unit.stdout.strip()[:35]))
                report.ok('>>ok\n')
    

    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close(cleanup=False)

    report.header()


    # fetch profiles and convert into inspectable data frames
    if not session:
        sys.exit(-1)


    import radical.pilot.utils as rpu

    # we have a session
    sid        = session.uid
    profiles   = rpu.fetch_profiles(sid=sid, tgt='/tmp/')
    profile    = rpu.combine_profiles (profiles)
    frame      = rpu.prof2frame(profile)
    sf, pf, uf = rpu.split_frame(frame)

  # print len(sf)
  # print len(pf)
  # print len(uf)
  # 
  # print sf[0:10]
  # print pf[0:10]
  # print uf[0:10]
    rpu.add_info(uf)

    for index, row in uf.iterrows():
        if str(row['info']) != 'nan':
            print "%-20s : %-10s : %-25s : %-20s" % \
                    (row['time'], row['uid'], row['state'], row['info'])

    rpu.add_states(uf)
    adv = uf[uf['event'].isin(['advance'])]
    print len(adv)
  # print uf[uf['uid'] == 'unit.000001']
  # print list(pf['event'])

    rpu.add_frequency(adv, 'f_exe', 0.5, {'state' : 'Executing', 'event' : 'advance'})
    print adv[['time', 'f_exe']].dropna(subset=['f_exe'])

    s_frame, p_frame, u_frame = rpu.get_session_frames(sid)
    print str(u_frame)

    for f in profiles:
        os.unlink(f)


#-------------------------------------------------------------------------------





import os
import zmq
import time
import errno
import pprint
import signal
import Queue           as pyq
import threading       as mt
import multiprocessing as mp
import radical.utils   as ru

# --------------------------------------------------------------------------
# defines for queue roles
QUEUE_INPUT   = 'input'
QUEUE_BRIDGE  = 'bridge'
QUEUE_OUTPUT  = 'output'
QUEUE_ROLES   = [QUEUE_INPUT, QUEUE_BRIDGE, QUEUE_OUTPUT]

# defines for queue types
QUEUE_THREAD  = 'thread'
QUEUE_PROCESS = 'process'
QUEUE_ZMQ     = 'zmq'
QUEUE_TYPES   = [QUEUE_THREAD, QUEUE_PROCESS, QUEUE_ZMQ]

_BRIDGE_TIMEOUT = 5.0  # how long to wait for bridge startup

# --------------------------------------------------------------------------
#
# zmq will (rightly) barf at interrupted system calls.  We are able to rerun
# those calls.
#
# FIXME: how does that behave wrt. tomeouts?  We probably should include
#        an explicit timeout parameter.
#
# kudos: https://gist.github.com/minrk/5258909
#
def _uninterruptible(f, *args, **kwargs):
    cnt = 0
    while True:
        cnt += 1
        try:
            return f(*args, **kwargs)
        except zmq.ZMQError as e:
            if e.errno == errno.EINTR:
                if cnt > 10:
                    raise
                # interrupted, try again
                print 'interrupted! [%s] [%s] [%s]' % (f, args, kwargs)
                continue
            else:
                # real error, raise it
                raise


# ==============================================================================
#
# Communication between components is done via queues.  The semantics we expect
# (and which is what is matched by the native Python Queue.Queue), is:
#
#   - multiple upstream   components put messages onto the same queue (input)
#   - multiple downstream components get messages from the same queue (output)
#   - local order of messages is maintained
#   - message routing is fair: whatever downstream component calls 'get' first
#     will get the next message (bridge)
#
# Additionally, we require Queues to be uni-directional, ie. Queues have an
# in-end for which one can call 'put()', and and out-end, for which one can call
# 'get()'.  This allows us to transparently (and easily) use different
# implementations of the communication channel, including zmq.
#
# The queue implementation we use depends on the participating component types:
# as long as we communicate within threads, Queue.Queue can be used.  If
# processes on the same host are involved, we switch to multiprocessing.Queue.
# For remote component processes we use zero-mq (zmq) queues.  In the cases
# where the setup is undetermined, we'll have to use zmq, too, to cater for all
# options.
#
# To make the queue type switching transparent, we provide a set of queue
# implementations and wrappers, which implement the interface of Queue.Queue:
#
#   put(msg)
#   get()
#   get_nowait()
#
# Not implemented is, at the moment:
#
#   qsize
#   empty
#   full 
#   put(msg, block, timeout)
#   put_nowait
#   get(block, timeout)
#   task_done
#   join
#
# Our Queue additionally takes 'name', 'role' and 'address' parameter on the
# constructor.  'role' can be 'input', 'bridge' or 'output', where 'input' is
# the end of a queue one can 'put()' messages into, and 'output' the end of the
# queue where one can 'get()' messages from. A 'bridge' acts as as a message
# forwarder.  'address' denominates a connection endpoint, and 'name' is
# a unique identifier: if multiple instances in the current process space use
# the same identifier, they will get the same queue instance.  Those parameters
# are mostly useful for the zmq queue.
#
class _QueueRegistry(object):

    __metaclass__ = ru.Singleton

    def __init__(self):

        # keep mapping between queue names and instances
        self._lock     = mt.RLock()
        self._registry = {QUEUE_THREAD  : dict(),
                          QUEUE_PROCESS : dict()}


    def get(self, flavor, name, ctor):

        if flavor not in QUEUE_TYPES:
            raise ValueError("no such type '%s'" % flavor)

        with self._lock:

            if name in self._registry[flavor]:
                # queue instance for that name exists - return it
              # print 'found queue for %s %s' % (flavor, name)
                return self._registry[flavor][name]

            else:
                # queue does not yet exist: create and register
                queue = ctor()
                self._registry[flavor][name] = queue
              # print 'created queue for %s %s' % (flavor, name)
                return queue

# create a registry instance
_registry = _QueueRegistry()


# ==============================================================================
#
class Queue(object):
    """
    This is really just the queue interface we want to implement
    """
    def __init__(self, flavor, qname, role, address=None):

        self._flavor = flavor
        self._qname  = qname
        self._role   = role
        self._addr   = address
        self._debug  = False
        self._name   = "queue.%s.%s" % (self._qname, self._role)
        self._log    = ru.get_logger('rp.bridges', target="%s.log" % self._name)

        if not self._addr:
            self._addr = 'tcp://*:*'

        if role in [QUEUE_INPUT, QUEUE_OUTPUT]:
            self._log.info("create %s - %s - %s - %s", flavor, qname, role, address)

    @property
    def name(self):
        return self._name

    @property
    def qname(self):
        return self._qname

    @property
    def flavor(self):
        return self._flavor

    @property
    def role(self):
        return self._role

    @property
    def addr(self):
        return self._addr


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Queue.
    #
    @classmethod
    def create(cls, flavor, name, role, address=None):

        # Make sure that we are the base-class!
        if cls != Queue:
            raise TypeError("Queue Factory only available to base class!")

        try:
            impl = {
                QUEUE_THREAD  : QueueThread,
                QUEUE_PROCESS : QueueProcess,
                QUEUE_ZMQ     : QueueZMQ,
            }[flavor]
          # print 'instantiating %s' % impl
            return impl(flavor, name, role, address)
        except KeyError:
            raise RuntimeError("Queue type '%s' unknown!" % flavor)


    # --------------------------------------------------------------------------
    #
    def poll(self):
        """
        check state of endpoint or bridge
        None: RUNNING
        0   : DONE
        1   : FAILED
        """
        return None


    # --------------------------------------------------------------------------
    #
    def put(self, msg):
        raise NotImplementedError('put() is not implemented')


    # --------------------------------------------------------------------------
    #
    def get(self):
        raise NotImplementedError('get() is not implemented')


    # --------------------------------------------------------------------------
    #
    def get_nowait(self):
        raise NotImplementedError('get_nowait() is not implemented')


    # --------------------------------------------------------------------------
    #
    def stop(self):
        raise NotImplementedError('stop() is not implemented')


# ==============================================================================
#
class QueueThread(Queue):

    def __init__(self, flavor, name, role, address=None):

        Queue.__init__(self, flavor, name, role, address)
        self._q = _registry.get(flavor, name, pyq.Queue)


    # --------------------------------------------------------------------------
    #
    def put(self, msg):

        if not self._role == QUEUE_INPUT:
            raise RuntimeError("queue %s (%s) can't put()" % (self._qname, self._role))

        self._q.put(msg)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get()" % (self._qname, self._role))

        return self._q.get()


    # --------------------------------------------------------------------------
    #
    def get_nowait(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get_nowait()" % (self._qname, self._role))

        try:
            return self._q.get_nowait()
        except pyq.Empty:
            return None


# ==============================================================================
#
class QueueProcess(Queue):

    def __init__(self, flavor, name, role, address=None):

        Queue.__init__(self, flavor, name, role, address)
        self._q = _registry.get(flavor, name, mp.Queue)


    # --------------------------------------------------------------------------
    #
    def put(self, msg):

        if not self._role == QUEUE_INPUT:
            raise RuntimeError("queue %s (%s) can't put()" % (self._qname, self._role))

        self._q.put(msg)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get()" % (self._qname, self._role))

        return self._q.get()


    # --------------------------------------------------------------------------
    #
    def get_nowait(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get_nowait()" % (self._qname, self._role))

        try:
            return self._q.get_nowait()
        except pyq.Empty:
            return None


# ==============================================================================
#
class QueueZMQ(Queue):


    def __init__(self, flavor, name, role, address=None):
        """
        This Queue type sets up an zmq channel of this kind:

            input \            / output
                   -- bridge -- 
            input /            \ output

        ie. any number of inputs can 'zmq.push()' to a bridge (which
        'zmq.pull()'s), and any number of outputs can 'zmq.request()' 
        messages from the bridge (which 'zmq.response()'s).

        The bridge is the entity which 'bind()'s network interfaces, both input
        and output type endpoints 'connect()' to it.  It is the callees
        responsibility to ensure that only one bridge of a given type exists.

        Addresses are of the form 'tcp://host:port'.  Both 'host' and 'port' can
        be wildcards for BRIDGE roles -- the bridge will report the in and out
        addresses as obj.bridge_in and obj.bridge_out.
        """

        self._p          = None           # the bridge process
        self._q          = None           # the zmq queue
        self._lock       = mt.RLock()     # for _requested
        self._requested  = False          # send/recv sync
        self._bridge_in  = None           # bridge input  addr
        self._bridge_out = None           # bridge output addr

        Queue.__init__(self, flavor, name, role, address)

        # ----------------------------------------------------------------------
        # behavior depends on the role...
        if self._role == QUEUE_INPUT:
            ctx = zmq.Context()
            self._q = ctx.socket(zmq.PUSH)
            self._q.connect(self._addr)


        # ----------------------------------------------------------------------
        elif self._role == QUEUE_BRIDGE:

            # we expect bridges to always use a port wildcard. Make sure
            # that's the case
            elems = self._addr.split(':')
            if len(elems) > 2 and elems[2] and elems[2] != '*':
                raise RuntimeError('wildcard port (*) required for bridge addresses (%s)' \
                                % self._addr)

            # ------------------------------------------------------------------
            def _bridge(addr, pqueue):

                try:
                    import setproctitle as spt
                    spt.setproctitle('radical.pilot %s' % self._name)
                except Exception as e:
                    pass

                try:
                    # reset signal handlers to their default
                    signal.signal(signal.SIGINT,  signal.SIG_DFL)
                    signal.signal(signal.SIGTERM, signal.SIG_DFL)
                    signal.signal(signal.SIGALRM, signal.SIG_DFL)

                    self._log.info('start bridge %s on %s', self._name, addr)

                    # FIXME: should we cache messages coming in at the pull/push 
                    #        side, so as not to block the push end?

                    ctx = zmq.Context()
                    _in = ctx.socket(zmq.PULL)
                    _in.bind(addr)

                    _out = ctx.socket(zmq.REP)
                    _out.bind(addr)

                    # communicate the bridge ports to the parent process
                    _in_port  =  _in.getsockopt(zmq.LAST_ENDPOINT)
                    _out_port = _out.getsockopt(zmq.LAST_ENDPOINT)

                    pqueue.put([_in_port, _out_port])

                    self._log.info('bound bridge %s to %s : %s', self._name, _in_port, _out_port)

                    # start polling for messages
                    _poll = zmq.Poller()
                    _poll.register(_out, zmq.POLLIN)

                    while True:

                        events = dict(_uninterruptible(_poll.poll, 1000)) # timeout in ms

                        if _out in events:
                            req = _uninterruptible(_out.recv)
                            _uninterruptible(_out.send_json, _uninterruptible(_in.recv_json))

                except Exception as e:
                    self._log.exception('bridge error: %s', e)
            # ------------------------------------------------------------------

            pqueue   = mp.Queue()
            self._p  = mp.Process(target=_bridge, args=[self._addr, pqueue])
            self._p.start()

            try:
                self._bridge_in, self._bridge_out = pqueue.get(True, _BRIDGE_TIMEOUT)
            except pyq.Empty as e:
                raise RuntimeError ("bridge did not come up! (%s)" % e)

        # ----------------------------------------------------------------------
        elif self._role == QUEUE_OUTPUT:
            ctx = zmq.Context()
            self._q = ctx.socket(zmq.REQ)
            self._q.connect(self._addr)

        # ----------------------------------------------------------------------
        else:
            raise RuntimeError ("unsupported queue role '%s'" % self._role)


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.stop()


    # --------------------------------------------------------------------------
    #
    @property
    def bridge_in(self):
        if self._role != QUEUE_BRIDGE:
            raise TypeError('bridge_in is only defined on a bridge')
        return self._bridge_in


    # --------------------------------------------------------------------------
    #
    @property
    def bridge_out(self):
        if self._role != QUEUE_BRIDGE:
            raise TypeError('bridge_out is only defined on a bridge')
        return self._bridge_out


    # --------------------------------------------------------------------------
    #
    def poll(self):
        """
        Only check bridges -- endpoints are otherwise always considered valid
        """
        if self._p and not self._p.is_alive():
            return 0


    # --------------------------------------------------------------------------
    #
    def stop(self):

        if self._p:
            self._p.terminate()


    # --------------------------------------------------------------------------
    #
    def put(self, msg):

        if not self._role == QUEUE_INPUT:
            raise RuntimeError("queue %s (%s) can't put()" % (self._qname, self._role))

      # self._log.debug("-> %s", pprint.pformat(msg))
        _uninterruptible(self._q.send_json, msg)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get()" % (self._qname, self._role))

        _uninterruptible(self._q.send, 'request')

        msg = _uninterruptible(self._q.recv_json)
      # self._log.debug("<- %s", pprint.pformat(msg))
        return msg


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout=None): # timeout in ms

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get_nowait()" % (self._qname, self._role))

        with self._lock: # need to protect self._requested

            if not self._requested:
                # we can only send the request once per recieval
                _uninterruptible(self._q.send, 'request')
                self._requested = True

          # try:
          #     msg = self._q.recv_json(flags=zmq.NOBLOCK)
          #     self._requested = False
          #     self._log.debug("<< %s", pprint.pformat(msg))
          #     return msg
          #
          # except zmq.Again:
          #     return None

            if _uninterruptible(self._q.poll, flags=zmq.POLLIN, timeout=timeout):
                msg = _uninterruptible(self._q.recv_json)
                self._requested = False
              # self._log.debug("<< %s", pprint.pformat(msg))
                return msg

            else:
                return None


# ------------------------------------------------------------------------------




#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.logentry
   :platform: Unix
   :synopsis: log entry wrapper

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import time

_iso = "%Y-%m-%dT%H:%M:%S.%f"

# ------------------------------------------------------------------------------
# The logentry "struct" encapsulates a log entry and its timestamp.
class Logentry(object):

    __slots__ = ('_message', '_timestamp')

    # --------------------------------------------------------------------------
    #
    def __init__(self, message, timestamp=None, logger=None):

        if not timestamp :
            timestamp = time.time()

        if  logger :
            logger (message)

        self._message   = message
        self._timestamp = timestamp


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def from_dict (d) :

        return Logentry (message=d['message'], timestamp=d['timestamp'])


    # --------------------------------------------------------------------------
    #
    @property
    def message(self):
        """Returns the message
        """
        return self._message

    # --------------------------------------------------------------------------
    #
    @property
    def timestamp(self):
        """Returns the timestamp
        """
        return self._timestamp

    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns the message and its timestamp as a Python dictionary.
        """
        return {"message"  : self.message, 
                "timestamp": self.timestamp}

    # --------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns the string representation of the log entry. The string 
           representation is just the state as string without its timestamp.
        """
        return self.message

# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import sys
import time
import radical.pilot as rp

# READ: The RADICAL-Pilot documentation: 
#   http://radicalpilot.readthedocs.org/en/latest
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if 
# you want to see what happens behind the scences!


#------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):
    """ this callback is invoked on all pilot state changes """

    print "[Callback]: ComputePilot '%s' state: %s." % (pilot.uid, state)


#------------------------------------------------------------------------------
#
def unit_state_cb (unit, state):
    """ this callback is invoked on all unit state changes """

    print "[Callback]: ComputeUnit  '%s' state: %s." % (unit.uid, state)


#------------------------------------------------------------------------------
#
def wait_queue_size_cb(umgr, wait_queue_size):
    """ 
    this callback is called when the size of the unit managers wait_queue
    changes.
    """
    print "[Callback]: UnitManager  '%s' wait_queue_size changed to %s." \
        % (umgr.uid, wait_queue_size)

    pilots = umgr.get_pilots ()
    for pilot in pilots:
        print "pilot %s: %s" % (pilot.uid, pilot.state)

    if wait_queue_size == 0:
        for pilot in pilots:
            if pilot.state in [rp.PENDING_LAUNCH,
                                rp.LAUNCHING     ,
                                rp.PENDING_ACTIVE]:
                print "cancel pilot %s" % pilot.uid
                umgr.remove_pilot (pilot.uid)
                pilot.cancel ()


#------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(name=session_name)
    print "session id: %s" % session.uid

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # prepare some input files for the compute units
        os.system ('hostname > file1.dat')
        os.system ('date     > file2.dat')
    
        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        pmgr = rp.PilotManager (session=session)
    
        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback (pilot_state_cb)
    
        # Define a 4-core local pilot that runs for 10 minutes and cleans up
        # after itself.
        pdesc1 = rp.ComputePilotDescription()
        pdesc1.resource = "localhost"
        pdesc1.runtime  = 10 # minutes
        pdesc1.cores    =  2
    
        pdesc2 = rp.ComputePilotDescription()
        pdesc2.resource = "localhost"
        pdesc2.runtime  = 10 # minutes
        pdesc2.cores    =  2
    
        # Launch the pilots
        pilots = pmgr.submit_pilots ([pdesc1, pdesc2])
    
        # wait for them to become active
        pmgr.wait_pilots (state=[rp.ACTIVE, rp.DONE, rp.FAILED])
        
    
        # Combine the ComputePilot, the ComputeUnits and a scheduler via
        # a UnitManager object.
        umgr = rp.UnitManager (session   = session,
                               scheduler = rp.SCHED_BACKFILLING)
    
        # Register our callback with the UnitManager. This callback will get
        # called every time any of the units managed by the UnitManager
        # change their state.
        umgr.register_callback (unit_state_cb, rp.UNIT_STATE)
    
        # Register also a callback which tells us when all units have been
        # assigned to pilots
        umgr.register_callback(wait_queue_size_cb,   rp.WAIT_QUEUE_SIZE)
    
    
        # Add the previously created ComputePilot to the UnitManager.
        umgr.add_pilots (pilots)
    
        # Create a workload of ComputeUnits (tasks). Each compute unit
        # uses /bin/cat to concatenate two input files, file1.dat and
        # file2.dat. The output is written to STDOUT. cu.environment is
        # used to demonstrate how to set environment variables within a
        # ComputeUnit - it's not strictly necessary for this example. As
        # a shell script, the ComputeUnits would look something like this:
        #
        #    export INPUT1=file1.dat
        #    export INPUT2=file2.dat
        #    /bin/cat $INPUT1 $INPUT2
        #
        cuds = []
        for unit_count in range(0, 32):
            cud = rp.ComputeUnitDescription()
            cud.executable    = "/bin/sh"
            cud.environment   = {'INPUT1': 'file1.dat', 'INPUT2': 'file2.dat'}
            cud.arguments     = ["-l", "-c", "cat $INPUT1 $INPUT2; sleep 10"]
            cud.cores         = 1
            cud.input_staging = ['file1.dat', 'file2.dat']
            cud.restartable   = True
    
            cuds.append(cud)
    
        # Submit the previously created ComputeUnit descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning ComputeUnits to the ComputePilots.
        units = umgr.submit_units(cuds)
    
        # the pilots have a total of 4 cores, and run for 10 min.  A CU needs about
        # 10 seconds, so we can handle about 24 units per minute, and need a total
        # of about 3 minutes.  We now wait for 60 seconds, and then cancel the first
        # pilot.  The 2 units currently running on that pilot will fail, all others
        # should get rescheduled to the other pilot.
        time.sleep (60)
        pilots[0].cancel ()
    
        # Wait for all compute units to reach a terminal state (DONE or FAILED).
        umgr.wait_units()
    
        print 'units all completed'
        print '----------------------------------------------------------------------'
    
        for unit in units:
            unit.wait ()
    
        for unit in units:
            print "* Task %s (executed @ %s) state %s, exit code: %s, started: %s, finished: %s, stdout: %s" \
                % (unit.uid, unit.execution_locations, unit.state, unit.exit_code, unit.start_time, unit.stop_time, unit.stdout)
    
        # delete the test data files
        os.system ('rm file1.dat')
        os.system ('rm file2.dat')

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print "caught Exception: %s" % e
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print "need to exit now: %s" % e

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print "closing session"
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


#-------------------------------------------------------------------------------



